from typing import Literal, Union, List, Optional


DATA_TYPE = Literal['Boolean', 'Number', 'String', 'DateTime', 'Integer', 'List']
SECTION_TYPE = Literal['Object', 'Array', 'StaticArray']

class FieldData:
    id: int
    label: str
    value: Union[bool, int, None, str]
    type: DATA_TYPE

    def __init__(self, id: int, label: str, value: Union[bool, int, None, str], type: DATA_TYPE) -> None:
        self.id = id
        self.label = label
        self.value = value
        self.type = type

class RowData:
    fields: List[FieldData]
    id: int

    def __init__(self, variables: List[FieldData], id: int) -> None:
        self.fields = variables
        self.id = id

    def getFieldById(self, id: int) -> FieldData:
        """Returns the field with the given id, or None if not found
        
        Parameters
        ----------
        id: int
            Field id setup in Form Editor

        Returns
        -------
        FieldData when found, None otherwise

        """

        for field in self.fields:
            if field.id == id:
                return field
        
        return None

    def getFieldByLabel(self, label: str) -> FieldData:
        """Returns the field with the given label, or None if not found
        
        Parameters
        ----------
        label: str
            Field label setup in Form Editor

        Returns
        -------
        FieldData when found, None otherwise

        """

        for field in self.fields:
            if field.label == label:
                return field
        
        return None

class SectionData:
    sectionId: int
    sectionName: str
    sectionType: SECTION_TYPE
    fields: Optional[List[FieldData]]
    rows: List[RowData]

    def __init__(self, id: int, name: str, type: SECTION_TYPE, fields: Optional[List[FieldData]], rows: List[RowData]) -> None:
        self.sectionId = id
        self.sectionName = name
        self.sectionType = type
        self.fields = fields
        self.rows = rows

    def getFieldById(self, id: int) -> FieldData:
        """Returns the field with the given id, or None if not found
        
        Parameters
        ----------
        id: int
            Field id setup in Form Editor

        Returns
        -------
        FieldData when found, None otherwise

        """

        for field in self.fields:
            if field.id == id:
                return field
        
        return None

    def getFieldByLabel(self, label: str) -> FieldData:
        """Returns the field with the given label, or None if not found
        
        Parameters
        ----------
        label: str
            Field label setup in Form Editor

        Returns
        -------
        FieldData when found, None otherwise

        """

        for field in self.fields:
            if field.label == label:
                return field
        return None

    def getRow(self, row_number: int) -> RowData:
        """Returns the row with the given row number, or None if not found
        
        Parameters
        ----------
        id: int
            Row id

        Returns
        -------
        RowData when found, None otherwise

        """

        for row in self.rows:
            if row.id == row_number:
                return row

        return None
    
class FormData:
    sections: dict[int, SectionData]

    def __init__(self, sections: List[SectionData]) -> None:
        self.sections = sections

    def getSectionById(self, section_id: int):
        """Returns the first section with the given id
        At present, section id is unchangeble in Form Editor so this function is safer than getSectionByName

        Parameters
        ----------
        section_id: UUID
            Section Id (order) of the section in Form Editor

        Returns
        -------
        SectionData when found, None otherwise

        """

        if section_id not in self.sections:
            return None

        return self.sections[section_id]

    def getSectionByName(self, section_name: str):
        """Returns the first section with the given name
        Caution, changing section name in Form Editor will break this function

        Parameters
        ----------
        section_name: str
            Name of the section setup in Form Editor

        Returns
        -------
        SectionData when found, None otherwise

        """

        for section in self.sections.values():
            if section.sectionName == section_name:
                return section

        return None


from typing import Dict, Literal


STATUS_STATE = Literal['Queued', 'Pending', 'Completed', 'ActionRequired', 'Failed']

STATUS_TYPE = Literal['Percentage',
                      'StateBased', 'Count', 'Fraction', 'Simple']

STEP_EVENTS = Literal['CheckStatusOnJourneyCreation',
                      'TriggerProcessOnStepCompletion']

SIGNALR_HUBNAMES = Literal['Journeys']
SIGNALR_TARGET_METHODS = Literal['journey-full']

JOURNEY_NOTIFICATION_FEATURE_NAME = 'platform-journeys'

JOURNEY_EVENT_DEFINITION_IDS = Literal['journey-completed',
                                        'journey-failed', 'journey-required']

JOURNEY_STEP_EVENT_DEFINITION_IDS = Literal['journey-step-completed',
                                            'journey-step-failed', 'journey-step-action-required']

JOURNEY_EVENT_DEFINITION_MAP: Dict[STATUS_STATE, JOURNEY_EVENT_DEFINITION_IDS] = {
    'Completed': 'journey-completed',
    'ActionRequired': 'journey-action-required',
    'Failed': 'journey-failed'
}

JOURNEY_STEP_EVENT_DEFINITIONS_MAP: Dict[STATUS_STATE, JOURNEY_STEP_EVENT_DEFINITION_IDS] = {
    'Completed': 'journey-step-completed',
    'ActionRequired': 'journey-step-action-required',
    'Failed': 'journey-step-failed'
}

NOTIFICATION_FIELD_PORTFOLIO_NAME = 'PortfolioName'



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


from enum import Enum
from typing import List, Optional, Union
from pydantic import BaseModel
from .literals import STATUS_STATE
from .step import SwitchGuideStep, SwitchGuideStepComponent, SwitchGuideStepDefinitionUiAssets, SwitchGuideStepDependency, SwitchGuideStepOverrides



class SwitchGuideScope(Enum):
    """Scope of the Guide"""

    PORTFOLIO = 'Portfolio'
    """Single guide instance for the entire Portfolio"""

    SITE = 'Site'
    """Single guide instance per Site in the Portfolio"""


class MarketplaceUserType(Enum):
    """User type for the Marketplace"""

    NONE = 'None'
    """Not available on the Marketplace
    Will be hidden from the marketplace if the guide was previously available.
    """

    ALL = 'All'
    """Available for all users"""

    SWITCHUSERONLY = 'SwitchUserOnly'
    """Available for Switch Users only
    Useful for testing a Guide before the guide is made available for all users in the Marketplace.
    """


class SwitchGuideStatus(BaseModel):
    state: STATUS_STATE
    percentageCompleted: int = 0


class SwitchGuideDefinitionOptions(BaseModel):
    enable_live_notification: bool = False
    """Enable live notification for this Guide"""

    enable_debug_mode: bool = False
    """Enable debug mode for switch user.
    Provides additional elements on the UI to help with debugging.
    Such as logs associated with each individual step.
    """

    availableOnMarketplaceForUserType: Optional[MarketplaceUserType] = MarketplaceUserType.ALL,
    """User types the guide is available for on the Marketplace"""


class SwitchGuideSummaryStepEvents(BaseModel):
    """Events that can be triggered on the Summary Step
    """

    componentOnCompletion: Union[SwitchGuideStepComponent, None]
    """Defined component will be displayed on the Summary Step when the Guide is completed to 100%"""


class SwitchGuideSummaryStepConfigDefinition(BaseModel):
    """Configuration for the Summary Step"""

    uiAssets: Optional[SwitchGuideStepDefinitionUiAssets]
    """UI Assets associated with the Guide"""

    events: Optional[SwitchGuideSummaryStepEvents]
    """Events that can be triggered on the Summary Step"""


class SwitchGuideSummaryStepConfig(BaseModel):
    component: Optional[SwitchGuideStepComponent]
    uiAssets: Optional[SwitchGuideStepDefinitionUiAssets]

class SwitchGuideDefinition(BaseModel):
    """Definition of the Guide"""
    
    id: str = ''
    """Unique identifier for the Guide"""

    name: str
    """Name of the Guide"""

    description: str
    """Description of the Guide"""

    instructions: str
    """Instructions on how to use the Guide"""

    imageUrl: Optional[str] = None
    """Image URL for the Guide
    This image will be displayed on the Marketplace.
    Relative to Switch Assets CDN example: static/hub-assets/example/location/of/image.png
    """

    tags: Optional[List[str]] = []
    """Tags associated with the Guide"""

    scope: Optional[SwitchGuideScope] = SwitchGuideScope.PORTFOLIO
    """Scope of the Guide"""

    summaryStep: Optional[SwitchGuideSummaryStepConfigDefinition]
    """Configuration for the Summary Step"""
    
    steps: List[SwitchGuideStepDependency]
    """Steps associated with the Guide"""
    
    options: Optional[SwitchGuideDefinitionOptions]
    """Options for the Guide"""


class SwitchGuideInstance(BaseModel):
    id: str
    status: SwitchGuideStatus
    steps: List[SwitchGuideStepOverrides]


class SwitchGuide(SwitchGuideDefinition, SwitchGuideInstance, BaseModel):
    id: str
    summaryStep: Optional[SwitchGuideSummaryStepConfig]
    steps: List[SwitchGuideStep]


class SwitchGuideSummary(BaseModel):
    id: str
    journeyDefinitionId: str
    name: str
    createdOnUtc: str
    modifiedOnUtc: str
    description: str
    instructions: str
    status: SwitchGuideStatus



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


from typing import List, Optional, Union
import uuid
from pydantic import BaseModel
from .guide import SwitchGuide, SwitchGuideDefinition, SwitchGuideInstance
from .step import SwitchGuideStepComponent, SwitchGuideStepDefinition, SwitchGuideStepDefinitionUiAssets, SwitchGuideStepData, SwitchGuideStepStatus, SwitchGuideStepStatusUiState


class SwitchGuideExecuteApiInput(BaseModel):
    journeyInstanceId: str
    journeyInstance: SwitchGuideInstance = None
    journeyDefinition: SwitchGuideDefinition = None
    journeyStepDefinitions: List[SwitchGuideStepDefinition] = []
    stepId: str = ''
    data: dict = ''


class SwitchGuideStepApiResponse(BaseModel):
    component: Union[SwitchGuideStepComponent, None] = None
    """Component to be displayed on the UI temporarily.
    Useful to display component to the user that's only relevant after a step is immediately executed.
    Example: Step execution fails and so we would like to display a error component temporarily with additional information.
    Note: This is not persisted in the database. Will be overwritten when check_ui_component is called.
    Ensure that check_ui_component is updated to match the component to display during a given state.
    """

    uiState: SwitchGuideStepStatusUiState = SwitchGuideStepStatusUiState()
    """Dynamically control UI state of the step.
    Maybe overwritten by check_status() function. Ensure they are in sync if experiencing issues.
    """

    data: dict = None
    """[Deprecated]
    Values set on this property are ignored. Will be removed in future versions.
    """

    errorMessage: str = ''
    """[Deprecated]
    Values set on this property are ignored. Will be removed in future versions.
    """

    status: SwitchGuideStepStatus = None
    """[Deprecated]
    Values set on this property are ignored. Will be removed in future versions.
    """

    uiAssets: SwitchGuideStepDefinitionUiAssets = None
    """[Deprecated]
    Values set on this property are ignored. Will be removed in future versions.
    """


class SwitchGuideApiResponse(BaseModel):
    success: bool = True
    errorMessage: str = ''
    journey: SwitchGuide = None
    journeyInstance: SwitchGuideInstance = None


class SwitchGuideFetchApiInput(BaseModel):
    journeyInstance: SwitchGuideInstance = None
    journeyDefinition: SwitchGuideDefinition = None
    journeyStepDefinitions: List[SwitchGuideStepDefinition] = []


class SwitchGuideFetchApiResponse(BaseModel):
    success: bool = True
    errorMessage: str = ''
    journey: SwitchGuide = None
    journeyInstance: SwitchGuideInstance = None


class SwitchGuideStepProcessInput(BaseModel):
    journey_id: uuid.UUID
    """[Deprecated] Prefer to use guide_id property as this will be removed in future versions."""

    guide_id: Optional[uuid.UUID]
    """Unique identifier for the Guide"""

    stepData: Optional[SwitchGuideStepData] = {}
    """Data associated with the step"""



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


class SwitchGuideDefinitionInvalidException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class InvalidStepApiReponseException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)


class InvalidStepStatusReponseException(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
Data Ingestion into the Switch Automation
=========================================

Complete package for ingestion data into the Switch Automation Platform.
"""

from . import models
from .models import api
from .models import step
from .models import guide
from .models import enums
from .models import literals

from . import exceptions
from . import tasks
from . import utils
from . import extensions
from .SwitchGuideTask import SwitchGuideTask

__all__ = [
    'models',
    'exceptions',
    'extensions',
    'tasks',
    'utils',
    'SwitchGuideTask',
    'api',
    'step',
    'guide',
    'literals',
    'enums'
]

__version__ = "0.1.17"



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


from abc import abstractmethod
from io import StringIO
import json
import logging
import sys
import pandas
import requests
from ..models.guide import SwitchGuideDefinition
import switch_api as sw
from ..utils.utils import ApiInputs, _column_name_cap, get_journeys_endpoint
from switch_api.pipeline.pipeline import Task

class GuideDefinitionTask(Task):
    logger = None

    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)
        consoleHandler = logging.StreamHandler(sys.stdout)
        consoleHandler.setLevel(logging.INFO)
        self.logger.handlers.clear()
        self.logger.addHandler(consoleHandler)
        formatter = logging.Formatter('%(asctime)s  switch_guides.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                                      datefmt='%Y-%m-%dT%H:%M:%S')
        consoleHandler.setFormatter(formatter)

    @property
    @abstractmethod
    def definition(self) -> SwitchGuideDefinition:
        """"Structure of the Journey"""
        return None

    def register(self, api_inputs: ApiInputs):
        if api_inputs == None or api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            self.logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        script_response = sw.pipeline.Automation.register_task(api_inputs, self)
        _, register = (0, script_response) if type(script_response) is pandas.DataFrame else script_response
        self.logger.info(register)

        headers = api_inputs.api_headers.default

        json_payload = {
            "scriptDriverId": str(self.id)
        }

        journey_endpoint = get_journeys_endpoint(api_inputs=api_inputs)
        
        url = f"{journey_endpoint}/projects/{api_inputs.api_project_id}/definitions/register"
        self.logger.info("Sending request: POST %s", url)

        response = requests.post(url, json=json_payload, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            self.logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                              response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            self.logger.error(
                'No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()
        
        response_data = pandas.read_json(StringIO(response.text), typ='Series').to_frame().T

        response_data.rename(columns = {'driverId':'guideDefinitionId'}, inplace = True)

        response_data.columns = _column_name_cap(response_data.columns)

        return response_status, response_data



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module defining extensions for Automation:
----------
Automation
----------
This class contains the helper methods extensions used to register, deploy, and test the created tasks. Additional helper
functions for retrieving details of existing tasks on the Switch Automation Platform are also included in this module.
"""
from .automation import deploy_as_upload_data_feed, deploy_on_timer
__all__ = ['deploy_as_upload_data_feed', 'deploy_on_timer']



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


from abc import abstractmethod
from io import StringIO
import logging
import sys
import pandas
import requests
import switch_api as sw

from ..exceptions import InvalidStepApiReponseException, InvalidStepStatusReponseException
from ..models.api import SwitchGuideStepApiResponse, SwitchGuideStepProcessInput

from ..models.step import SwitchGuideStepComponent, SwitchGuideStepDefinition, SwitchGuideStepDefinitionUiAssets, SwitchGuideStepStatus
from ..utils.utils import ApiInputs, _column_name_cap, get_journeys_endpoint
from switch_api.pipeline.pipeline import Task

class GuideStepDefinitionTask(Task):
    logger = None

    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.DEBUG)
        consoleHandler = logging.StreamHandler(sys.stdout)
        consoleHandler.setLevel(logging.INFO)
        self.logger.handlers.clear()
        self.logger.addHandler(consoleHandler)
        formatter = logging.Formatter('%(asctime)s  switch_guides.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                                      datefmt='%Y-%m-%dT%H:%M:%S')
        consoleHandler.setFormatter(formatter)
    
    @property
    @abstractmethod
    def definition(self) -> SwitchGuideStepDefinition:
        """"Structure of the Journey Step"""
        return None

    @abstractmethod
    def check_status(self, api_inputs: ApiInputs, journey_input: SwitchGuideStepProcessInput) -> SwitchGuideStepStatus:
        """This method will be called to understand current status of the step.
        Ideally the method is very quick and limits calls to retrieve data only for building current status. 

        Args:
            api_inputs (ApiInputs): Provides state for calling Switch Platform APIs.

        Returns:
            SwitchGuideStepStatus: response containing the status of the step or None
        """
        return None
    
    @abstractmethod
    def check_ui_assets(self, step_status: SwitchGuideStepStatus) -> SwitchGuideStepDefinitionUiAssets:
        """This method will be called to understand what would be the ui assets based on the status.
        Method can return empty response, None, or call Super method if not required.
        Ideally this method should not involve any external calls. 

        Args:
            step_status (SwitchGuideStepStatus): The journey step status to evaluate.

        Returns:
            SwitchGuideStepDefinitionUiAssets: response containing UiAsset overrides.
        """
        return None
    
    @abstractmethod
    def check_ui_component(self, api_inputs: ApiInputs, step_status: SwitchGuideStepStatus, journey_input: SwitchGuideStepProcessInput) -> SwitchGuideStepComponent:
        """This method is called to return any component overrides
        Method can return empty response, None, or call Super method if not required.
        Ideally this method should not involve any external calls. 

        Args:
            api_inputs (ApiInputs): Provides state for calling Switch Platform APIs.
            step_status (SwitchGuideStepStatus): The journey step status to evaluate.
            journey_input (SwitchGuideStepProcessInput): Information specific to a Journey

        Returns:
            SwitchGuideStepComponent: response containing ui component override or None
        """
        return None

    @abstractmethod
    def process(self, api_inputs: ApiInputs, journey_input: SwitchGuideStepProcessInput) -> SwitchGuideStepApiResponse:
        """ This method requires user wait for completion. 
        How long this method should run depends on how long a user is willing to wait on the page. 
        Use GuideStepDefinitionAsyncTask if you want to run this method in background.

        This method will be called by the Guides Backend under two conditions.
            1. When user completes the requirements of a component associated with this step.
        
        Add your logic here to extract, process, and export data how ever you see fit.
        """
        return None
    
    def update_component(self, response: SwitchGuideStepApiResponse, component: SwitchGuideStepComponent):
        if (response == None or type(response) is not SwitchGuideStepApiResponse):
            self.logger.error('Provided response model is not a valid')
            raise InvalidStepApiReponseException('Expected valid SwitchGuideStepApiResponse')
        
        response.component = component
        
    def update_status(self, response: SwitchGuideStepApiResponse, status: SwitchGuideStepStatus):
        if (response == None or type(response) is not SwitchGuideStepApiResponse):
            self.logger.error('Provided response model is not a valid')
            raise InvalidStepApiReponseException('Expected SwitchGuideStepApiResponse')
        
        if (status == None):
            self.logger.error('Provided Step Status model is not a valid')
            raise InvalidStepStatusReponseException('Expected valid SwitchGuideStepStatus')
        
        if (status.type != self.definition.status.type):
            self.logger.error('Step Status type cannot be changed via script. Please update the definition model.')
            raise InvalidStepStatusReponseException('Cannot update Step Status type via script')
        
        response.status = status

    def register(self, api_inputs: ApiInputs):
        if api_inputs == None or api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
            self.logger.error("You must call initialize() before using API.")
            return pandas.DataFrame()

        script_response = sw.pipeline.Automation.register_task(api_inputs, self)

        _, register = (0, script_response) if type(script_response) is pandas.DataFrame else script_response

        self.logger.info(register)

        if (register.size == 0):
            self.logger.error(
                "Journey Definition registration was unsuccessful")
            return pandas.DataFrame()

        headers = api_inputs.api_headers.default

        json_payload = {
            "scriptDriverId": str(self.id)
        }

        journey_endpoint = get_journeys_endpoint(api_inputs=api_inputs)

        url = f"{journey_endpoint}/projects/{api_inputs.api_project_id}/definitions/steps/register"
        self.logger.info("Sending request: POST %s", url)

        response = requests.post(url, json=json_payload, headers=headers)
        response_status = '{} {}'.format(response.status_code, response.reason)
        if response.status_code != 200:
            self.logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                              response.reason)
            return response_status, pandas.DataFrame()
        elif len(response.text) == 0:
            self.logger.error(
                'No data returned for this API call. %s', response.request.url)
            return response_status, pandas.DataFrame()

        response_data = pandas.read_json(StringIO(response.text), typ='Series').to_frame().T
        response_data.rename(columns = {'driverId':'guideStepDefinitionId'}, inplace = True)
        response_data.columns = _column_name_cap(response_data.columns)

        return response_status, response_data


# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


from abc import abstractmethod
from typing import List, Union
from .utils.utils import ApiInputs
import uuid
from switch_api.pipeline.pipeline import Task

from .models.api import SwitchGuideExecuteApiInput, SwitchGuideApiResponse, SwitchGuideFetchApiInput, SwitchGuideFetchApiResponse, SwitchGuideStepApiResponse
from .models.guide import SwitchGuide, SwitchGuideInstance, SwitchGuideDefinition, SwitchGuideStatus, SwitchGuideSummaryStepConfig
from .models.step import SwitchGuideStep, SwitchGuideStepDefinition, SwitchGuideStepOverrides, SwitchGuideStepStatus


class SwitchGuideTask(Task):
    """Switch Guide Task

    This class is used to run Switch Guide within the Switch Automation Platform.

    """

    DEFAULT_INDEX: int = -1

    def _find_step_in_journeyDef(self, definition: SwitchGuideDefinition, stepId: str):
        index = -1
        for step in definition.steps:
            index = index + 1
            if (uuid.UUID(step.stepId, version=4) == uuid.UUID(stepId, version=4)):
                return (step, index)

        return (None, self.DEFAULT_INDEX)

    def _find_step_def(self, steps: List[SwitchGuideStepDefinition], stepId: str):
        index = self.DEFAULT_INDEX
        for step in steps:
            index = index + 1
            if (uuid.UUID(step.stepId, version=4) == uuid.UUID(stepId, version=4)):
                return (step, index)

        return (None, self.DEFAULT_INDEX)

    def _find_step_in_journey(self, stepId: str, journeyInstance: SwitchGuideInstance):
        index = self.DEFAULT_INDEX
        steps: List[SwitchGuideStepOverrides] = journeyInstance.steps
        for step in steps:
            index = index + 1
            if (step.stepId == stepId):
                return (step, index)

        return (None, self.DEFAULT_INDEX)

    def build_journey_instance(self, definition: SwitchGuideDefinition, journeyInstanceId: str, journeyStepDefinitions: List[SwitchGuideStepDefinition]):
        journeySteps = []

        for s in definition.steps:
            (stepDefinition, _) = self._find_step_def(
                journeyStepDefinitions, s.stepId)

            if (stepDefinition == None):
                continue

            journeySteps.append(SwitchGuideStepOverrides(
                stepId=s.stepId,

                status=SwitchGuideStepStatus(
                    type=stepDefinition.status.type,
                    progress=stepDefinition.status.progress,
                    state=stepDefinition.status.state,
                    uiState=stepDefinition.status.uiState
                )
            ))

        journeyInstance = SwitchGuideInstance(
            # This should come from request and should not be definitionId as it's for an instance of definition
            id=journeyInstanceId,
            status=SwitchGuideStatus(
                # TODO: Check based on current state of steps
                percentageCompleted=0,
                # TODO: Check based on current state of steps or current step fails with ActionRequired or Failed
                state='Pending'
            ),
            steps=journeySteps
        )

        return journeyInstance

    def execute(self, api_inputs: ApiInputs, journey_input: SwitchGuideExecuteApiInput) -> SwitchGuideApiResponse:
        """ DEPRECATED: Method is called by the Journey framework to execute a specific step

        Args:
            api_inputs (ApiInputs): ApiInputs for calling SwitchApi methods
            journey_input (JourneyInput): Input for running Journey step

        Raises:
            Exception: thrown when step response is unexpected 

        Returns:
            JourneyResponse: response associated with the exceution of a step
        """

        currentStepId = journey_input.stepId

        # Future: Retrieved from the Database. Should be passed in from SwitchGuideExecuteApiInput.
        definition = journey_input.journeyDefinition

        stepDefinitions = journey_input.journeyStepDefinitions

        journeyInstance = journey_input.journeyInstance
        if (journeyInstance == None):
            journeyInstance = self.build_journey_instance(
                definition=definition, journeyInstanceId=journey_input.journeyInstanceId, journeyStepDefinitions=stepDefinitions)

        # Future: Retrieved from the Database when Step Definitions are stored there.
        (currentStepDef, index) = self._find_step_def(
            stepDefinitions, currentStepId)

        if (index == self.DEFAULT_INDEX):
            return SwitchGuideApiResponse(success=False, errorMessage="Step Definition could not be found")

        (journeyInstanceStep, _) = self._find_step_in_journey(
            currentStepId, journeyInstance)

        # Future: CurrentStepDef needs to be updated with values from JourneyStepOverrides so latest is available for step.
        result = self.handle_step(
            api_inputs=api_inputs, stepDef=currentStepDef)

        if (type(result) is not SwitchGuideStepApiResponse):
            raise Exception(
                f'Expected SwitchGuideStepApiResponse but got {type(result)}')

        journeySteps = self._build_journey_steps(
            definition=definition, journeyInstance=journeyInstance)

        # Update Journey Steps with changes made during handle_step
        for s in journeySteps:
            if (s.stepId != currentStepId):
                continue

            # TODO: update on journeyStep and journeyInstanceStep are duplicated

            journeyInstanceStep.status.state = result.status.state
            s.status.state = result.status.state

            journeyInstanceStep.status.progress = result.status.progress
            s.status.progress = result.status.progress

            if (result.status.messages):
                journeyInstanceStep.status.messages = result.status.messages
                s.status.messages = result.status.messages

            if (result.uiAssets):
                # TODO: This should be a little smarter and only override values in uiAssets that have changed.
                journeyInstanceStep.uiAssets = result.uiAssets
                s.uiAssets = result.uiAssets

        # Build finalized model of the Journey
        (journey, journeyInstance) = self._build_journey(
            definition=definition, journeyInstance=journeyInstance, journeySteps=journeySteps)

        return SwitchGuideApiResponse(journey=journey, journeyInstance=journeyInstance)

    def get_journey(self, journeyInput: SwitchGuideFetchApiInput) -> SwitchGuideApiResponse:
        definition = journeyInput.journeyDefinition

        journeyInstance = journeyInput.journeyInstance

        steps = journeyInput.journeyStepDefinitions

        journeySteps = self._build_journey_steps(
            definition=definition, journeyInstance=journeyInstance, journeyStepDefinitions=steps)

        (journey, journeyInstance) = self._build_journey(definition=definition,
                                                         journeyInstance=journeyInstance, journeySteps=journeySteps)

        return SwitchGuideFetchApiResponse(journey=journey, journeyInstance=journeyInstance)

    def _build_journey_steps(self, definition: SwitchGuideDefinition,
                             journeyInstance: SwitchGuideInstance, journeyStepDefinitions: List[SwitchGuideStepDefinition]) -> List[SwitchGuideStep]:

        journeySteps: List[SwitchGuideStep] = []

        # Build Journey Steps
        for step in definition.steps:
            (stepDefinition, index) = self._find_step_def(
                journeyStepDefinitions, step.stepId)

            if (index == self.DEFAULT_INDEX):
                continue

            (journeyDefinitionStep, _) = self._find_step_in_journeyDef(
                definition, step.stepId)

            (journeyInstanceStep, _) = self._find_step_in_journey(
                step.stepId, journeyInstance)

            if (journeyInstanceStep == None):
                # Step doesn't exist in Journey Instance.
                # Might happen when a step is added to Journey Definition
                #   after creation of Journey Instance
                journeyInstanceStep = SwitchGuideStepOverrides(
                    stepId=step.stepId,
                    status=SwitchGuideStepStatus(
                        type=stepDefinition.status.type,
                        progress=stepDefinition.status.progress,
                        state=stepDefinition.status.state
                    )
                )

                journeyInstance.steps.append(journeyInstanceStep)
            else:
                # Ensure Status Type is set according to Step Definition
                journeyInstanceStep.status.type = stepDefinition.status.type

                # Override progress component of the Step definition with one from Journey Instance Step
                stepDefinition.status.progress = journeyInstanceStep.status.progress
                stepDefinition.status.state = journeyInstanceStep.status.state
                stepDefinition.status.uiState = journeyInstanceStep.status.uiState

                if (journeyInstanceStep.uiAssets):
                    stepDefinition.uiAssets = journeyInstanceStep.uiAssets

            journeySteps.append(SwitchGuideStep(
                stepId=stepDefinition.stepId,
                callToAction=stepDefinition.callToAction,
                component=stepDefinition.component,
                description=stepDefinition.description,
                icon=stepDefinition.icon,
                isEnabled=stepDefinition.isEnabled,
                isHidden=stepDefinition.isHidden,
                name=stepDefinition.name,
                uiAssets=stepDefinition.uiAssets,
                status=stepDefinition.status,
                dependsOn=journeyDefinitionStep.dependsOn,
                order=journeyDefinitionStep.order,
                events=journeyDefinitionStep.events,
                isBackgroundStep=journeyDefinitionStep.isBackgroundStep
            ))

        return journeySteps

    def _build_journey(self, definition: SwitchGuideDefinition, journeyInstance: SwitchGuideInstance, journeySteps: List[SwitchGuideStep]) -> Union[SwitchGuide, SwitchGuideInstance]:
        # Update Journey Steps with overrides from JourneyInstance
        for s in journeySteps:
            (journeyInstanceStep, _) = self._find_step_in_journey(
                s.stepId, journeyInstance)

            s.status.state = journeyInstanceStep.status.state
            s.status.progress = journeyInstanceStep.status.progress
            s.status.uiState = journeyInstanceStep.status.uiState

            if (journeyInstanceStep.status.messages):
                s.status.messages = journeyInstanceStep.status.messages

            if(journeyInstanceStep.uiAssets):
                s.uiAssets = journeyInstanceStep.uiAssets

            if(journeyInstanceStep.component):
                s.component = journeyInstanceStep.component
            
            if(journeyInstanceStep.data):
                s.data = journeyInstanceStep.data

        journeyInstance = self.update_journey_instance_status(journeyInstance, journeySteps)

        # Represents the entire Journey combining Definition and Instance views
        # Future: Construction of SwitchGuide model should be done on the AF so we can fetch latest model from SQL before updating it.
        #   We could only update the specific step status rather than upserting the entire SwitchGuideInstance json document.
        #   In other words, we need to handle multiple accessing different steps at the same time.
        #   We could consider preventing updates to the journey if the user is working off an outdated state.
        #   We can provide an option for the user to refresh the journey when this happens.
        journey = SwitchGuide(
            id=journeyInstance.id,
            name=definition.name,
            description=definition.description,
            instructions=definition.instructions,
            status=journeyInstance.status,
            steps=journeySteps,  # Combine JourneyDef + JourneyInstance Steps
            summaryStep=self._get_summary_step(journey_definition=definition, journey_instance=journeyInstance)
        )

        return (journey, journeyInstance)

    def _get_summary_step(self, journey_definition: SwitchGuideDefinition, journey_instance: SwitchGuideInstance):
        if journey_instance.status.state != 'Completed':
            return None
        
        if not journey_definition.summaryStep or not journey_definition.summaryStep.events.componentOnCompletion:
            return None
        
        return SwitchGuideSummaryStepConfig(
            component=journey_definition.summaryStep.events.componentOnCompletion
        )

    def update_journey_instance_status(self, journeyInstance: SwitchGuideInstance, journeySteps: List[SwitchGuideStep]):
        # Update Journey Status
        completedSteps = [
            s for s in journeySteps if s.status.state == 'Completed']
        
        stepContributionRatio = 1 / len(journeySteps)

        if (len(completedSteps) == len(journeySteps)):
            journeyInstance.status.state = 'Completed'
            journeyInstance.status.percentageCompleted = 100
        else:
            journeyInstance.status.state = 'Pending'
            journeyInstance.status.percentageCompleted = int(
                len(completedSteps) / len(journeySteps) * 100)
            
        for s in journeySteps:
            if s.status.state == 'Completed':
                continue
            
            if s.status.type == 'Percentage':
                journeyInstance.status.percentageCompleted += int(stepContributionRatio * s.status.progress.completionPercentage)
            elif s.status.type == 'Fraction' and s.status.progress.denominator > 0:
                journeyInstance.status.percentageCompleted += int(stepContributionRatio * (s.status.progress.numerator / s.status.progress.denominator * 100))
                
        if (journeyInstance.status.percentageCompleted > 100):
            journeyInstance.status.percentageCompleted = 100
            
        return journeyInstance

    @abstractmethod
    def handle_step(self, api_inputs: ApiInputs, stepDef: SwitchGuideStepDefinition) -> SwitchGuideStepApiResponse:
        """Method to be implemented to process Guides Journey steps

        The method should contain all code used to execute steps associated specified within JourneyInput

        Parameters
        ----------
        api_inputs: ApiInputs
            object returned by call to initialize()
        journey_input : JourneyInput
            Any settings required to be passed to run Switch Guide steps. Step to process, query keys etc.
        """
        return SwitchGuideStepApiResponse()



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


from typing import Optional, Union
from typing import List
from pydantic import BaseModel
from .literals import STATUS_STATE, STATUS_TYPE


class SwitchGuideStepComponent(BaseModel):
    type: str
    id: str
    attributes: dict


class SwitchGuideStepCallToAction(BaseModel):
    path: str
    step: Optional[str] = ''
    description: Optional[str] = ''
    autoRedirectOnStepCompletion: Optional[bool] = False


class SwitchGuideStepStatusProgress(BaseModel):
    """Contains properties that help build the full picture of the status of a step. Not all properties might be relevant for each step.

    Args:
        BaseModel (_type_): Helper type for managing validations and serialization of the model
    """
    count: int = 0
    completionPercentage: int = 0
    numerator: int = 0
    denominator: int = 0


class SwitchGuideStepData(dict):
    pass


class SwitchGuideStepStatusUiState(BaseModel):
    """Keep the step locked on the UI when True; otherwise unlock it.
    Attribute relevant for UI
    """

    lockStep: bool = False
    """Keep the step locked on the UI when True; otherwise unlock it."""

    returnToSummary: bool = False
    """Indicates that we should return user to Summary step on Step Execution.
    Note: This will take the user to the even if next step is available.
    Note 2: Supersedes the 'continueToNextStep' property.
    """

    continueToNextStep: bool = False
    """Indicates that we should continue to next step on Step Execution.
    Note: This will take the user to the next step even when next step is already completed.
    Note 2: This property is ignored if the 'returnToSummary' property is set to True.
    """

    continueToNextStepWhenAvailable: bool = False
    """Indicates that we should continue to next step on Step Execution when next is available.
    Note: This will take the user to the next available incomplete step otherwise it will take the user to summary step.
    Note 2: This property is ignored if the 'returnToSummary' property is set to True.
    Note 3: This property is ignored if the 'continueToNextStep' property is set to True.
    """

class SwitchGuideStepStatus(BaseModel):
    """Status of the Journey Step
    """

    type: STATUS_TYPE = 'Percentage'
    """Type of status this is. Based on the type different property values would be relevant

    Available types and their relevant properties:

    Percentage
        - progress.percentageCompleted    [int]
        - messages.default                [str]

    StateBased
        - messages.completed      [str]   (when state == Completed)
        - messages.pending        [str]   (when state == Pending)
        - messages.actionRequired [str]   (when state == ActionRequired)
        - messages.failed         [str]   (when state == Failed)

    Count
        - progress.count [int]
        - messages.count [str]  (replaces occurrences of {{count}} in the message with the value of progress.count)

    Fraction
        - progress.numerator   [int]
        - progress.denominator [int]
        - messages.default     [str] (replaces occurrences of {{numerator}} and {{denominator}} in the message with the values of progress.numerator and progress.denominator respectively)

    Simple
        - messages.default     [str]

    """

    state: STATUS_STATE = 'Queued'
    """State of the step"""

    messages: dict = {}
    """Messages associated with the step. Contains at least 'default' message"""

    progress: SwitchGuideStepStatusProgress = SwitchGuideStepStatusProgress()
    """Progress data"""

    uiState: SwitchGuideStepStatusUiState = SwitchGuideStepStatusUiState()
    """UI State
    Dynamically control UI state of the step.
    """

class SwitchGuideStepDefinitionUiAssets(BaseModel):
    """Dynamically adjustable config to change UI look and feel. 
    """
    
    stepContinueButtonText: str = 'Get Started'
    """Step continue button text
    """

    progressCtaDescription: Optional[str] = None
    """Call To Action description to associate with the path.
    """

    progressCtaPath: Optional[str] = None
    """Call To Action path. Expected to be an internal route in the Platform App for now.
    """

    progressCtaText: Optional[str] = 'Click here'
    """Call To Action text. By default it will be "Click here".
    """

    stepBannerText: Optional[str] = None
    """Text contained in the banner.
    """

    stepBannerType: Optional[str] = None
    """Type of the banner. By default it will be of Type 'Note'.
    In the future it could also be warnings and errors.
    """

    isPromotional: Optional[bool] = False
    """Treat step as a promotional step with limited function.
    Purely for cosmetic purposes. Specific limitations should be handled by other means.
    """

    enableUpgrade: Optional[bool] = False
    """Controls whether to display Upgrade request button across the Guide.
    """

class SwitchGuideStepDefinition(BaseModel):
    """Shape of the a Step. This defines what a step is and the data is shared by instances of the step.

    Args:
        BaseModel (_type_): Helper type for managing validations and serialization of the model
    """
    
    stepId: str = ''
    """Id of the Step - same as the Step Definition driver id."""

    name: str
    """Name of the step visible to the user on the Guides UI"""
    
    description: str
    """Description visible to the user on the Guides UI"""

    status: SwitchGuideStepStatus = SwitchGuideStepStatus()
    """Status associated with the Step"""
    
    icon: str
    """Icon associated with the step"""

    isEnabled: bool = True
    """Indicates whether the step is enabled or not. If not enabled, the step will not be visible to the user."""

    isHidden: bool = False
    """When True, hides the step on the left tracker of the Guides UI. 
    And displays in the summary page when (if any) onDepends conditions are met.
    When False, step will always be visible on the left tracker and summary page of the Guides UI.
    """

    callToAction: Union[SwitchGuideStepCallToAction, None] = None
    
    component: Union[SwitchGuideStepComponent, None] = None
    """Custom component to display on the Guides UI
    """

    uiAssets: SwitchGuideStepDefinitionUiAssets = SwitchGuideStepDefinitionUiAssets()
    """Configuration to customize certain elements of the Guides UI
    """


class SwitchGuideStepDependencyEvents(BaseModel):
    checkStatusOnJourneyCreation: bool = True
    """Check status of the step on Journey creation"""

    triggerProcessOnStepCompletion: List[str] = []
    """List of step ids that should be processed when this step is completed"""

    preventStatusCheckBeforeDependsOn: bool = False
    """When true, status check will not be called for this step until dependent steps are completed"""


class SwitchGuideStepDependency(BaseModel):
    """Represents the instance of a step where the state might not be the same as other instances of the same step
    """

    stepId: str
    """Id of the step instance"""

    order: Optional[int]
    """Order to appear on the Frontend UIs.
        Optional when when Step is marked as a Background Step
    """
    
    dependsOn: List[str] = []
    """When a step depends on another step, the step is only unlocked on UI when the other step is completed.
    """
    
    events: SwitchGuideStepDependencyEvents = SwitchGuideStepDependencyEvents()
    """Events associated with the step in relation to the journey or other steps within the journey.
    """

    isBackgroundStep: bool = False
    """When True, step is expected to run on event trigger and will not be returned to the frontend.
    """

    options: Optional[dict] = {}
    """Options to be passed to the step driver. Options are unique to each step so consult the Step documentation."""

class SwitchGuideStepOverrides(BaseModel):
    """Represents the simplest view of a step. Useful for maintain overrides for the step without duplicating the data.

    Args:
        BaseModel (_type_): Helper type for managing validations and serialization of the model
    """

    stepId: str
    status: SwitchGuideStepStatus
    component: Union[SwitchGuideStepComponent, None] = None
    uiAssets: Optional[SwitchGuideStepDefinitionUiAssets]
    data: Optional[SwitchGuideStepData] = {}


class SwitchGuideStep(SwitchGuideStepDefinition, SwitchGuideStepDependency):
    """Full Journey Step Model

    Args:
        SwitchGuideStepDefinition (_type_): Represents the step shape
        SwitchGuideStepDependency (_type_): Represents the shape of a step within a Journey dependency
    """
    stepId: str
    data: Optional[SwitchGuideStepData] = {}



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


from . import literals
from . import api
from . import step
from . import guide
from . import enums

__all__ = [
    'api',
    'step',
    'guide',
    'literals',
    'enums'
]



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import json
from typing import List, Union
import uuid
from switch_api._utils._utils import ApiInputs
import requests
import switch_api as sw
from ...models.step import SwitchGuideStepComponent
from .form_data import FormData, SectionData, FieldData, RowData


def define_ui_component(form_id: str, form_keys: List[str] = [], options = [], data = []) -> SwitchGuideStepComponent:
    """
    Defines a basic Forms Viewer Component to be displayed on the UI.

    Parameters
    ----------
    form_id: UUID
        Form Definition Id
    form_keys: List[str]
        Optional, list of form keys used to scope data to a specific form instance
    options: List[dict]
        Optional, list of configuration options
    data: List[dict]
        Optional, list of data prepopulate the form

    Returns SwitchGuideStepComponent
    -------
    """

    component = SwitchGuideStepComponent(
        type='Vue',
        id='FormsViewer',
        attributes={
            'formId': form_id,
            # Set in check_ui_component()
            'formDataKeys': [],
            'configurationOptions': [
                {
                    "section": 1,
                    "fields": [],
                    "rows": []
                }
            ],
            'configData': [
                {
                    "section": 1,
                    "fields": [],
                    "rows": []
                }
            ]
        }
    )

    if form_keys and len(form_keys) > 0:
        component.attributes['formDataKeys'] = form_keys

    if options and len(options) > 0:
        component.attributes['configurationOptions'] = options

    if data and len(data) > 0:
        component.attributes['configData'] = data

    return component


def get_data(api_inputs: ApiInputs, form_id: uuid.UUID, form_key: str = '') -> Union[FormData, None]:
    """
    Retrieves the data from a form instance in the Switch platform.

    Parameters
    ----------
    api_inputs: ApiInputs
        Object returned by initialize() function.
    form_id: UUID
        Form Definition Id
    form_key_value: str
        Optional, form key used to form scope data to a specific form instance
        When not provided, will scope data to the Portfolio

    Returns
    -------
    """

    headers = api_inputs.api_headers.integration

    if not form_key:
        form_key = str(uuid.UUID(int=0))
    
    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/jsonforms/{form_id}/keyvalue/{form_key}"
    response = requests.get(url, headers=headers)

    response_status = '{} {}'.format(response.status_code, response.reason)

    if response.status_code != 200:
        sw.pipeline.logger.error(f"Error retrieving form data: {response_status}")
        return None
    elif len(response.text) == 0:
        sw.pipeline.logger.error(f"Empty form data retrieved: {response_status}")
        return None

    response_json = json.loads(response.text)

    form_data = _extract_form_data(response_json)

    return form_data

def _extract_form_data(form_data) -> FormData:
    section_data: dict[int, SectionData] = {}

    if len(form_data) == 0:
        return None

    try:
        for section in form_data['data']['data']:
            section_data[section['id']] = SectionData(
                id=section['id'],
                name=section['name'],
                type=section['type'],
                fields=_extract_fields(section['data']),
                rows=_extract_rows(section['data'])
            )
    except:
        sw.pipeline.logger.error(f"Form data extraction failed", exc_info=True)

    form = FormData(sections=section_data)

    return form

def _extract_fields(fields: List[dict]) -> List[FieldData]:
    field_data: List[FieldData] = []

    for field in fields['variables']:
        field_data.append(FieldData(
            id=field['id'],
            label=field['label'],
            value=field['value'],
            type=field['type']
        ))

    return field_data

def _extract_rows(rows: List[dict]) -> List[RowData]:
    row_data: List[RowData] = []

    for row in rows['data']:
        row_data.append(RowData(
            variables=_extract_fields(row),
            id=row['id']
        ))

    return row_data



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


from . import utils
from .utils import ApiInputs

__all__ = [
    'utils',
    'ApiInputs'
]



from collections import namedtuple

ApiInputs = namedtuple('ApiInputs',
                       ['email_address', 'user_id', 'api_project_id', 'data_feed_id', 'data_feed_file_status_id',
                        'bearer_token', 'api_base_url', 'api_projects_endpoint', 'subscription_key', 'api_headers'])

def get_journeys_endpoint(api_inputs: ApiInputs):
    return api_inputs.api_base_url.replace("task-insights/api/1.0", "journeys")
    # return 'http://localhost:8000'

def coalesce(*values):
    """Return the first non-None value or None if all values are None"""
    return next((v for v in values if v is not None), None)

def _column_name_cap(columns) -> list:
    renamed_columns = [name[0].upper() + name[1:] for name in columns.to_list()]
    return renamed_columns


# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# Import required functions
from setuptools import setup, find_packages

# Call setup function
setup(
    author="Switch Automation Pty Ltd.",
    description="A package for building Platform Guides in Switch Automation Platform.",
    long_description=open('README.md', 'r').read() + '\n\n' + open('HISTORY.md', 'r').read(),
    long_description_content_type='text/markdown',
    license='MIT License',
    name="switch_guides",
    version="0.1.17",
    packages=find_packages(
        include=["switch_guides", "switch_guides.*"],
        exclude=["switch_guides.tests", "switch_guides.tests.*"]
    ),
    install_requires=['switch-api','pydantic==1.9.0'],
    python_requires=">=3.8.*",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        "License :: OSI Approved :: MIT License",
        'Intended Audience :: Other Audience',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Natural Language :: English',
    ]
)



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module defining the Automation extensions which contains deployment.
---------------------
Automation Extensions
---------------------
This class contains the helper methods extensions used to register, deploy, and test the created tasks. Additional helper functions
 for retrieving details of existing tasks on the Switch Automation Platform are also included in this module.
"""
import logging
import sys
import pandas
import requests
import uuid
from io import StringIO
from switch_api._utils._utils import _column_name_cap
from ..utils.utils import ApiInputs
from switch_api._utils._constants import (
    EXPECTED_DELIVERY, QUEUE_NAME, SCHEDULE_TIMEZONE)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
consoleHandler = logging.StreamHandler(sys.stdout)
consoleHandler.setLevel(logging.INFO)
logger.handlers.clear()
logger.addHandler(consoleHandler)
formatter = logging.Formatter('%(asctime)s  switch_api.%(module)s.%(funcName)s  %(levelname)s: %(message)s',
                              datefmt='%Y-%m-%dT%H:%M:%S')
consoleHandler.setFormatter(formatter)


def deploy_as_upload_data_feed(driver_id: str,
                               api_inputs: ApiInputs, data_feed_id: uuid.UUID, expected_delivery: EXPECTED_DELIVERY,
                               queue_name: QUEUE_NAME = 'task', data_feed_name: str = None):
    """Deploy the custom driver as a REST API end point Datafeed.
    To upload a file to the deployed data feed, use the UploadUrl from the response dataframe (with request type
    POST) with the following two headers:
    - 'Ocp-Apim-Subscription-Key' - set to the value of ``api_inputs.subscription_key``
    - 'Authorization' - set to the value 'Bearer ``api_inputs.bearer_token``'
    For example, to upload a file using the ``requests`` package:
    >>> import requests
    >>> url = df.loc[0,'UploadUrl']
    >>> payload={}
    >>> file_path = 'C:/xxyyzz.txt'
    >>> files={'file': open(file_path, 'rb')}
    >>> headers = {'Ocp-Apim-Subscription-Key': api_inputs.subscription_key, 'Authorization': f'Bearer {api_inputs.bearer_token}'}
    >>> response = requests.request("POST", url, headers=headers, data=payload, files=files)
    >>> print(response.text)
    Parameters
    ----------
    taskId : str
        Id of the custom driver class created from the Abstract Base Class 'Task'
    api_inputs : ApiInputs
        Object returned by the initialize() function.
    data_feed_id : uuid.UUID
        The DataFeedId to update if existing, else will create a new record with the given value.
    expected_delivery : EXPECTED_DELIVERY
        The expected delivery frequency of the data.
    queue_name : QUEUE_NAME, optional
        The queue name (Default value = 'task').
    data_feed_name : str, Optional
        The name of the data feed (to be displayed in Task Insights UI).
    Returns
    -------
    df : pandas.DataFrame
        Dataframe containing the details of the deployed https endpoint data feed.
    """
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()
    if data_feed_name == '' and data_feed_name == None:
        logger.error('Data Feed Name must be provided.')
        return pandas.DataFrame()
    headers = api_inputs.api_headers.default
    if not set([expected_delivery]).issubset(set(EXPECTED_DELIVERY.__args__)):
        logger.error('expected_delivery parameter must be set to one of the allowed values defined by the '
                     'EXPECTED_DELIVERY literal: %s', EXPECTED_DELIVERY.__args__)
        return pandas.DataFrame()
    if not set([queue_name]).issubset(set(QUEUE_NAME.__args__)):
        logger.error('queue_name parameter must be set to one of the allowed values defined by the '
                     'QUEUE_NAME literal: %s', QUEUE_NAME.__args__)
        return pandas.DataFrame()
    logger.info(
        f'Deploy Driver with Id {driver_id} as a data feed for ApiProjectID: {api_inputs.api_project_id}')
    payload = {
        "dataFeedId": str(data_feed_id),
        "driverId": str(driver_id),
        "name": data_feed_name,
        "feedType": 'Readings',  # Default
        "expectedDelivery": expected_delivery,
        "sourceType": "Upload",
        "queueName": queue_name,
        "upload": {
            "placeholder": ""
        },
    }
    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/tasks/deployment"
    logger.info("Sending request: POST %s", url)
    response = requests.post(url, json=payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error(
            'No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()
    df = pandas.read_json(StringIO(response.text),
                          typ='Series').to_frame().T
    df.columns = _column_name_cap(df.columns)
    return response_status, df


def deploy_on_timer(driver_id: str,
                    api_inputs: ApiInputs, data_feed_id: uuid.UUID, expected_delivery: EXPECTED_DELIVERY,
                    cron_schedule: str, queue_name: QUEUE_NAME = "task", settings: dict = None,
                    schedule_timezone: SCHEDULE_TIMEZONE = 'Local', timezone_offset_minutes: int = None,
                    data_feed_name: str = None):
    """Deploy driver to run on timer.
    Parameters
    ----------
    task : Task
        The custom driver class created from the Abstract Base Class `Task`.
    api_inputs : ApiInputs
        Object returned by initialize.initialize() function
    data_feed_id : uuid.UUID
        The DataFeedId to update if existing, else will create a new record with the given value.
    expected_delivery : EXPECTED_DELIVERY
        The expected delivery frequency.
    cron_schedule : str
        The CRONOS cron object containing the required schedule for the driver to be run. For details on the
        required format, see: https://crontab.cronhub.io/
    queue_name : QUEUE_NAME, optional
        The queue name (Default value = 'task').
    settings : dict, Optional
        List of settings used to deploy the driver. For example, may contain the user_name and password required to
        authenticate calls to a third-party API (Default value = None).
    schedule_timezone : SCHEDULE_TIMEZONE, optional
        Whether the ``cron_schedule`` should be applied based on Local or Utc timezone. If set to `Local`, this is
        taken as the timezone of the western-most site in the given portfolio (Default value = 'Local').
    timezone_offset_minutes: int, Optional
        Timezone offset in minutes (from UTC) to be used when applying the ``cron_schedule`` (Default value = None).
    data_feed_name : str, Optional
        The name of the data feed (to be displayed in Task Insights UI). If not provided, will default to the
        task name.
    Returns
    -------
    pandas.Dataframe
        A dataframe containing the details of the deployed data feed.
    """
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()
    if data_feed_name == '' and data_feed_name == None:
        logger.error('Data Feed Name must be provided.')
        return pandas.DataFrame()
    if timezone_offset_minutes is None:
        timezone_offset_minutes = 0
    headers = api_inputs.api_headers.default
    if not set([expected_delivery]).issubset(set(EXPECTED_DELIVERY.__args__)):
        logger.error('expected_delivery parameter must be set to one of the allowed values defined by the '
                     'EXPECTED_DELIVERY literal: %s', EXPECTED_DELIVERY.__args__)
        return pandas.DataFrame()
    if not set([schedule_timezone]).issubset(set(SCHEDULE_TIMEZONE.__args__)):
        logger.error('schedule_timezone parameter must be set to one of the allowed values defined by the '
                     'SCHEDULE_TIMEZONE literal: %s', SCHEDULE_TIMEZONE.__args__)
        return pandas.DataFrame()
    if not set([queue_name]).issubset(set(QUEUE_NAME.__args__)):
        logger.error('queue_name parameter must be set to one of the allowed values defined by the '
                     'QUEUE_NAME literal: %s', QUEUE_NAME.__args__)
        return pandas.DataFrame()
    if 5 > len(cron_schedule.split(' ')) > 6:
        logger.error(
            "cron_schedule parameter must be in the format * * * * *")
        return pandas.DataFrame()
    logger.info(
        f'Deploy for Driver Id {driver_id} on timer for ApiProjectID: {api_inputs.api_project_id} and schedule: {cron_schedule}.')
    logger.info(
        f'Settings to be passed to the driver on start are: {str(settings)}')
    payload = {
        "dataFeedId": str(data_feed_id),
        "driverId": driver_id,
        "name": data_feed_name,
        "feedType": "Readings",  # Default
        "expectedDelivery": expected_delivery,
        "sourceType": "Timer",
        "queueName": queue_name,
        "timer": {
            "cronSchedule": cron_schedule,
            "timezoneOffsetMinutes": timezone_offset_minutes,
            "scheduleTimezone": schedule_timezone
        },
        "settings": settings
    }
    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/tasks/deployment"
    logger.info("Sending request: POST %s", url)
    response = requests.post(url, json=payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)
    if response.status_code != 200 and len(response.text) > 0:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        error_df = pandas.read_json(response.text)
        return response_status, error_df
    elif response.status_code != 200 and len(response.text) == 0:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error(
            'No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()
    df = pandas.read_json(response.text, typ='Series').to_frame().T
    df.columns = _column_name_cap(df.columns)
    return df


def reschedule_timer(api_inputs: ApiInputs, data_feed_id: uuid.UUID, queue_on_schedule: bool = False):
    """Rescedule Timer driver.
    Parameters
    ----------
    api_inputs : ApiInputs
        Object returned by initialize.initialize() function
    data_feed_id : uuid.UUID
        The DataFeedId of the Timer Task to reschedule
    Returns
    -------
    pandas.Dataframe
        A dataframe containing the details of the deployed data feed.
    """
    if api_inputs.api_base_url == '' or api_inputs.bearer_token == '':
        logger.error("You must call initialize() before using API.")
        return pandas.DataFrame()

    headers = api_inputs.api_headers.default
    payload = {
        "dataFeedId": str(data_feed_id),
        "queueJobOnSchedule": queue_on_schedule
    }

    url = f"{api_inputs.api_projects_endpoint}/{api_inputs.api_project_id}/data-ingestion/tasks/reschedule"
    logger.info("Sending request: POST %s", url)
    response = requests.post(url, json=payload, headers=headers)
    response_status = '{} {}'.format(response.status_code, response.reason)

    if response.status_code != 200 and len(response.text) > 0:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        error_df = pandas.read_json(response.text)
        return response_status, error_df
    elif response.status_code != 200 and len(response.text) == 0:
        logger.error("API Call was not successful. Response Status: %s. Reason: %s.", response.status_code,
                     response.reason)
        return response_status, pandas.DataFrame()
    elif len(response.text) == 0:
        logger.error(
            'No data returned for this API call. %s', response.request.url)
        return response_status, pandas.DataFrame()

    df = pandas.read_json(response.text, typ='Series').to_frame().T
    df.columns = _column_name_cap(df.columns)

    return df



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module defining extensions for Automation:
----------
Automation
----------
This class contains the helper methods extensions used to register, deploy, and test the created tasks. Additional helper
functions for retrieving details of existing tasks on the Switch Automation Platform are also included in this module.
"""
from .GuideDefinitionTask import GuideDefinitionTask
from .GuideRetrySupportTask import GuideRetrySupportTask
from .GuideStepDefinitionTask import GuideStepDefinitionTask
from .GuideStepDefinitionBackgroundTask import GuideStepDefinitionBackgroundTask

__all__ = ['GuideDefinitionTask', 'GuideRetrySupportTask', 'GuideStepDefinitionTask', 'GuideStepDefinitionBackgroundTask']



# -------------------------------------------------------------------------
# Copyright (c) Switch Automation Pty Ltd. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


from abc import abstractmethod, ABC
from ..utils.utils import ApiInputs

class GuideRetrySupportTask(ABC):
    @property
    @abstractmethod
    def check_retry(self, api_inputs: ApiInputs) -> bool:
        """"True when Task should be reprocessed; otherwise False"""
        return False
